//---------------------------------------------------------------------------------------
// Датчик RMS на Arduino Pro Mini 5V/16MHz
// Файл rms_17.04.10
// OldBean
//---------------------------------------------------------------------------------------
// Краткое описание процесса синхронизации задач.
//---------------------------------------------------------------------------------------
// Параллельно выполняются следующие задачи:
// 1. Периодическая оцифровка входного сигнала (выпрямленное диодным мостиком напряжение сети,
//    пониженное до необходимого уровня трансформатором напряжения и резистивным делителем).
// 2. По окончанию преобразования производится возведение результата в квадрат и суммирование
//    с предыдущими значениями. После накопления необходимого для усреднения количества
//    результатов, производится деление суммы на количество просуммированных результатов,
//    извлечение квадратного корня и нормировка.
// 3. Параллельно с процессами 1 и 2 производится регенерация трезразрядного семичегментного
//    индикатора для визуализации среднеквадратичного напряжения сети.
//
// Для устранения коллизий все эти процессы жестко синхронизированы. Тактирование работы 
// производится таймером TIMER2 - каждые 200 мкс происходят прерывания по совпадению c OCR2A.
// В процедуре обработки этого прерывания производится запуск АЦП. Частота преобразования АЦП
// установлена на 125 кГц, что вполне достаточно для 10-битного преобразования. Время 
// преобразования колеблется в пределах от 100 до 120 мкс (округленно). 
// 
// После 15 прерываний таймера (т.е. приблизительно раз в 3 мс) производится вызов функции 
// ShowSegments(), осуществляющей переход к подстветке следующей группы сегментов. Время 
// работы этой функции составляет около 28 мкс. Таким образом, переключение подсветки сегментов 
// происходит параллельно с циклом преобразования АЦП и никогда не выходит за его пределы.
// 
// По окончанию цикла преобразования АЦП происходит прерывание от АЦП. До начала следующего цикла
// преобразования АЦП остается приблизительно 80 - 100 мкс. За это время, в процедуре обработки
// этого прерывания от таймера, производится возведение результата оцифровки в квадрат и 
// суммирование этих квадратов. Здесь же, после суммирования 2048 значений (это соответствует 
// 41 полупериоду колебаний сети 50 Гц) производится деление деление суммы на 2048 (сдвиг 
// вправо на 11 бит) и извлечение квадратного корня из результата деления. Корень извлекается
// за время порядка 42 мкс, что вполне укладывается во временной интервал, остающийся до начала
// следующего цикла преобразования АЦП.
//---------------------------------------------------------------------------------------
#include <Wire.h>
#define SLAVE_ADDRESS       0x05      // Адрес шины i2c для связи с внешним миром. Slave
//---------------------------------------------------------------------------------------
// Процедура калибровки
//---------------------------------------------------------------------------------------
// Для калибровки требуется образцовый вольтметр. Если форма напряжения сети не сильно 
// отличается от синусоидальной, то в качестве образцового вольтметра может быть использован
// любой электронный или стрелочный вольтметр, измеряющий действующее (среднеквадратичное)
// значение. Если напряжения сети сильно отличается от синусоидального, то в качестве
// образцового вольметра нужно использовать, так называемые, TrueRMS - вольтметры.
// 
// Для калибровки необходимо выполнить следующие шаги.
//
// 1. Раскомментировать строку //#define CALIBR. При этом нормировочные коэффициенты будут
// такими, что на индикаторе будет высвечиваться код, соответствующий действующему напряжению
// сети.
// 2. Компиллировать и загрузить скетч в Arduino Pro Mini
// 3. Подключить датчик RMS к сети через ЛАТР. Параллельно подключить образцовый вольтметр RMS.
// 4. Установить на ЛАТРе напряжение, соответствующее верхней (максимальной) границе рабочего
// диапазона.
// 5. Регулируя делитель (резистор R5 на схеме) добиться чтобы код, высвечиваемый на индикаторе
// датчика, не превышал 600-650 единиц.
// 6. Регулируя напряжение ЛАТРом и контролируя напряжение образцовым вольтметром, убедиться 
// в линейности датчика во всем рабочем диапазоне.
// 7. Измерить напряжение и коды на границах рабочего диапазона и подставить их значения в
// соответствующие позиции строки:
// static int U1 = xxx, C1 = xxx, U2 = xxx, C2 = xxx;
// где U1, C1 - напряжение и код на нижней границе рабочего диапазона, а U2, C2 - соответственно,
// на верхней.
// 8. Закомментировать строку #define CALIBR
// 9. Компиллировать и загрузить скетч
//---------------------------------------------------------------------------------------
//#define DEBUG
//---------------------------------------------------------------------------------------
//#define CALIBR                   // Раскомментировать эту строчку при калибровке
//---------------------------------------------------------------------------------------
static int N = 50;
static int N2 = N*N; // Количество отсчетов (в сумме) для вычисления RMS
#ifdef CALIBR
static int kS = 0, kM = 1, kD = N; // Коэффициенты для выполнения калибровки
#else
static long U1 = 140, C1 = 336, U2 = 260, C2 = 650; // Напряжения и коды на границах диапазона
static long kS = C2*U1*N - C1*U2*N, kM = U2 - U1, kD = C2*N - C1*N;
#endif
//---------------------------------------------------------------------------------------
#define COMMON_CATHODE // Если индикатор с общим анодом - необходимо закомментировать эту строку
//---------------------------------------------------------------------------------------
#ifdef COMMON_CATHODE // Для индикаторов с общим катодом
#define L_HIGH 1
#define L_LOW  0
#else                 // Для индикаторов с общим анодом
#define L_HIGH 0
#define L_LOW  1
#endif
//---------------------------------------------------------------------------------------
static uint8_t adcChannel = 3; // Входной аналоговый сигнал подключен к пину A3
//---------------------------------------------------------------------------------------
// Установки, связанные с индикатором, рассчитаны на трехразрядные семисегментные индикаторы
// с общими катодами (или анодами). Выводы анодов (или катодов для индикаторов с общими анодами) 
// подключены непосредственно к соответствующим пинам МК. Выводы общих катодов (общих анодов для
// индикаторов с общими анодами) подключены через резисторы 470 Ом.
//---------------------------------------------------------------------------------------
// Пины, к которым подключены сегменты индикатора - аноды (или катоды для индикаторов с общими анодами)
static uint8_t segPins[7] = {12, 16, 5, 7, 8, 13, 4};
// Пины, к которым подключены общие катоды (или аноды, для индикаторов с общими анодами)
static uint8_t digPins[3] = {15, 14, 11};
static uint8_t digit[11] = {0b00111111, // Маски сегментов цифр 0, 1, 2 и т.д.
                            0b00000110,
                            0b01011011,
                            0b01001111,
                            0b01100110,
                            0b01101101,
                            0b01111101,
                            0b00000111,
                            0b01111111,
                            0b01101111,
                            0b00000000}; // "Цифра" 10 это пробел
static int d[3]; // Сотни, десятки и единицы показываемые на разрядах 7-сегментного индикатора
static uint8_t mask = 0x01; // Маска для анализа цифр
static int seg = 0; // Номер сегмента, обрабатываемого (подствечиваемого, если нужно) в данный момент
static volatile int cntr = 0; // Счетчик для тактирования смены подствечиваемых сегментов

static volatile int rms = 137; // Среднеквадратичное значение измеряемого напряжения
static volatile unsigned long sum = 0; // Сумматор квадратов напряжения
static volatile int scntr = 0; // Счетчик квадратов прибавленных к сумме
//---------------------------------------------------------------------------------------
void UpdateDigits(int val); // Подготовка данных (цифр) для визуализации
void ShowSegments(); // Подсвечивание очередной группы сегментов
void receiveData(int byteCount); // callback для принимаемых по i2c байтов
void sendData(); // callback для передаваемых по i2c байтов
//---------------------------------------------------------------------------------------
void setup() {
  Wire.begin(SLAVE_ADDRESS); // Инициализация библиотеки для шины i2c
  Wire.onReceive(receiveData);
  Wire.onRequest(sendData);

  for(int i = 0; i < 7; i++) { // Инициализация пинов, к которым подключены сегменты индикатора
    pinMode(segPins[i], OUTPUT);
    digitalWrite(segPins[i], L_LOW);
  }
  for(int i = 0; i < 3; i++) { // Инициализация пинов, к которым подключены общие катоды (аноды) индикатора
    pinMode(digPins[i], OUTPUT);
    digitalWrite(digPins[i], L_HIGH);
  }
  unsigned long st = millis() + 1000;
  d[0] = 8; d[1] = 8; d[2] = 8; // Для демонстрации работоспособности индикатора покажем все сегменты разом
  while(millis() < st) {
      ShowSegments();
  }
  d[0] = 10; d[1] = 10; d[2] = 10; // Очистим индикатор
  ShowSegments();

#ifdef DEBUG
  pinMode(9, OUTPUT); // 9 - это PB1. Здесь используется для измерения временных характеристик МК
  digitalWrite(9, LOW);
//  PORTB |= 0b00000010; // Фронт импульса на пине 9
// .... фрагмент кода, длительность выполнения которого измеряется .....
//  PORTB &= 0b11111101; // Спад импульса на пине 9
#endif

  // Инициализация АЦП
  ADMUX = (1 << REFS1) | (1 << REFS0) |  // 11 - опорное напряжение = 1.1В
          (adcChannel & 0b00001111); // Канал АЦП соответствует пину, к которому поключен входной сигнал
  ADCSRA = (1 << ADEN) | // АЦП включен
           (0 << ADATE) | // Режим однократного преобразования
           (1 << ADIE) | // Прерывания по окончанию преобразования разрешены
           (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0); // Предделитель - 128

  // Инициализация таймера 2 для тактирования АЦП и посветки индикатора
  TCCR2A = (1<<WGM21);            // Режим CTC (сброс по совпадению)
  TCCR2B = (1<<CS21) | (1<<CS20); // Предделитель 32
  OCR2A = 98;                     // При этом значении совпадения происходят с частотой около 5 кГц.
  TIMSK2 = (1<<OCIE2A);           // Разрешить прерывание по совпадению
}
//---------------------------------------------------------------------------------------
ISR (TIMER2_COMPA_vect) { // Обработчик прерывания от таймера TIMER2
  ADCSRA |= (1 << ADSC); // Запуск АЦП
  if(++cntr >= 15) { // Устанавливаем флаг для перехода к посветке следующей группы сегментов
    cntr = 0; ShowSegments(); // Подсвечиваем очередную группу сегментов
  }
}
//---------------------------------------------------------------------------------------
ISR(ADC_vect) { // Обработчик перрывания от АЦП по окончаниии преобразования
  int adcVal = ADCL; adcVal += ADCH << 8;
  sum += (unsigned long)adcVal*adcVal;
  if(++scntr == N2) { // Накопление квадратов напряжения закончено
    unsigned long wrk = sqrt(sum); // Извлекаем корень
#ifdef DEBUG
    PORTB |= 0b00000010;
#endif
    wrk *= kM; wrk += kS; wrk /= kD; // Нормировка
#ifdef DEBUG
    PORTB &= 0b11111101;
#endif
    sum = 0; scntr = 0; rms = wrk;
    UpdateDigits(rms);
  }
}
//---------------------------------------------------------------------------------------
void loop() {  
}
//---------------------------------------------------------------------------------------
void UpdateDigits(int val) { // Подготовка данных (цифр) для визуализации
  d[2] = val/100;
  d[1] = (val - d[2]*100)/10;
  d[0] = val - d[2]*100 - d[1]*10;
  if(d[2] == 0) { // Не отображаем ведущие нули ежели что... ("Цифра" 10 - это "пустая" цифра, все ее сегменты потушены)
    d[2] = 10;
    if(d[1] == 0) d[1] = 10;
  }
}
//---------------------------------------------------------------------------------------
void ShowSegments() { // Подсвечивание очередной группы сегментов
  // Комментарии написаны для индикаторов с общими катодами. В скобках - для индикаторов с общими анодами
  digitalWrite(segPins[seg], L_LOW); // Опускаем (поднимаем) потенциал анода (катода) "отработавшего" сегмента
  if(++seg < 7) mask <<= 1; // Переходим к следующему сегменту
  else { // Дошли до последнего сегмента
    seg = 0; mask = 0x01;
  }
  digitalWrite(segPins[seg], L_HIGH); // Поднимаем (роняем) потенциал анода (катода) нового текущего сегмента
  for(int bi = 0; bi < 3; bi++) { // Теперь пробегаем по цифрам индикатора
    if(digit[d[bi]] & mask) // Если данный сегмент в bi-ом разряде должен гореть,
      digitalWrite(digPins[bi], L_LOW); // то опустим (поднимем) потенциал катода (анода) соответствующей цифры
    else // А если гореть не должен, то
      digitalWrite(digPins[bi], L_HIGH); // потенциал общего катода (анода) bi-й цифры должен быть высокий (низкий)
  }
}
//---------------------------------------------------------------------------------------
void receiveData(int byteCount) { // callback для принимаемых по i2c байтов
  while(Wire.available()) {
    Wire.read(); // Просто затычка
  }
}
//---------------------------------------------------------------------------------------
void sendData() { // callback для передаваемых по i2c байтов
  uint8_t wrk = (rms > 100)? rms - 100: 0;
  Wire.write(wrk); // rms изменяется в диапазоне от 140 до 280. Для того, чтобы не передавать
                         // два байта, rms просто уменьшается на 100. Это необходимо учесть
                         // принимающей стороне.
}
//---------------------------------------------------------------------------------------
// 17.01.18.
// 1. Исправлена функция sedData, отправлявшая некорректное значение при rms < 100
//
// 17.04.10
// 1. Вставлены объявления (прототипы) функций в начале скетча

